/*
	BASSenc cast source example
	Copyright (c) 2006-2022 Un4seen Developments Ltd.
*/

#import "ViewController.h"
#include "bass.h"
#include "bassenc.h"

#ifdef USE_ADDONS // use BASSenc add-ons instead of command-line encoders
// these headers should be placed alongside bassenc.h
#include "bassenc_mp3.h"
#include "bassenc_ogg.h"
#include "bassenc_opus.h"
#include "bassenc_flac.h"
#endif

@implementation ViewController

HRECORD rchan;		// recording/encoding channel
HENCODE encoder;	// encoder handle
int reconwait;
int titlewait;

const DWORD bitrates[14] = { 32,40,48,56,64,80,96,112,128,160,192,224,256,320 }; // available bitrates

// display error messages
void Error(NSString *es)
{
	es = [es stringByAppendingFormat:@"\n(error code: %d)",BASS_ErrorGetCode()];
	if (![NSThread isMainThread]) {
		dispatch_async(dispatch_get_main_queue(), ^{
			NSAlert *alert=[[NSAlert alloc] init];
			[alert setMessageText:es];
			[alert runModal];
		});
	} else {
		NSAlert *alert=[[NSAlert alloc] init];
		[alert setMessageText:es];
		[alert runModal];
	}
}

BOOL32 CALLBACK RecordingCallback(HRECORD handle, const void *buffer, DWORD length, void *user)
{
	return BASS_Encode_IsActive(handle); // continue recording if encoder is alive
}

// encoder death notification
void CALLBACK EncoderNotify(HENCODE handle, DWORD status, void *user)
{
	if (status < 0x10000) { // encoder died
		dispatch_async(dispatch_get_main_queue(), ^{
			ViewController *view=(__bridge ViewController*)user;
			if (view.reconnectSwitch.state) {
				reconwait = 10; // trigger auto-reconnect
			} else {
				[view Stop]; // free the recording and encoder
				NSAlert *alert=[[NSAlert alloc] init];
				[alert setMessageText:status == BASS_ENCODE_NOTIFY_CAST ? @"The server connection died!" : @"The encoder died!"];
				[alert runModal];
			}
		});
	}
}

- (void)Start {
	char buf[200];
	const char *server, *pass, *name, *url, *genre, *desc, *content;
	DWORD codec, bitrate, flags;
	// start recording @ 44100hz 16-bit stereo (paused to setup encoder first)
	if (!(rchan = BASS_RecordStart(44100, 2, BASS_RECORD_PAUSE, RecordingCallback, 0))) {
		Error(@"Couldn't start recording");
		return;
	}
	codec = (int)self.codecSelector.indexOfSelectedItem;
	bitrate = bitrates[(int)self.bitrateSelector.indexOfSelectedItem]; // get bitrate
	// start the encoder
	if (codec == 4) { // AAC
		encoder = BASS_Encode_StartCA(rchan, 'adts', 'aac ', BASS_ENCODE_AUTOFREE, bitrate * 1000, 0, 0);
		content = BASS_ENCODE_TYPE_AAC;
	} else {
#ifdef USE_ADDONS
		if (!codec) { // MP3
			sprintf(buf, "-b %d", bitrate);
			encoder = BASS_Encode_MP3_Start(rchan, buf, BASS_ENCODE_AUTOFREE, 0, 0);
			content = BASS_ENCODE_TYPE_MP3;
		} else if (codec == 1) { // Ogg Vorbis
			sprintf(buf, "-M %d -m %d", bitrate, bitrate);
			encoder = BASS_Encode_OGG_Start(rchan, buf, BASS_ENCODE_AUTOFREE, 0, 0);
			content = BASS_ENCODE_TYPE_OGG;
		} else if (codec == 2) { // Opus
			sprintf(buf, "--bitrate %d --cvbr --max-delay 100 --padding 0", bitrate);
			encoder = BASS_Encode_OPUS_Start(rchan, buf, BASS_ENCODE_AUTOFREE, 0, 0);
			content = BASS_ENCODE_TYPE_OGG;
		} else { // FLAC
			sprintf(buf, "--ogg --no-padding --no-seektable --limit-min-bitrate");
			encoder = BASS_Encode_FLAC_Start(rchan, buf, BASS_ENCODE_AUTOFREE, 0, 0);
			content = BASS_ENCODE_TYPE_OGG;
			bitrate = 0; // undefined bitrate
		}
#else
		// setup encoder command-line (raw PCM data to avoid length limit)
		if (!codec) { // MP3
			sprintf(buf, "lame -b %d --flush -r -s 44100 -", bitrate); // add "-x" for LAME versions pre-3.98
			content = BASS_ENCODE_TYPE_MP3;
		} else if (codec == 1) { // Ogg Vorbis
			sprintf(buf, "oggenc -M %d -m %d -r -R 44100 -", bitrate, bitrate);
			content = BASS_ENCODE_TYPE_OGG;
		} else if (codec == 2) { // Opus
			sprintf(buf, "opusenc --bitrate %d --cvbr --max-delay 100 --padding 0 --raw --raw-rate 44100 - -", bitrate);
			content = BASS_ENCODE_TYPE_OGG;
		} else { // FLAC
			sprintf(buf, "flac --ogg --no-padding --no-seektable --limit-min-bitrate --sample-rate=44100 --channels=2 --bps=16 --endian=little --sign=signed -");
			content = BASS_ENCODE_TYPE_OGG;
			bitrate = 0; // undefined bitrate
		}
		encoder = BASS_Encode_Start(rchan, buf, BASS_ENCODE_NOHEAD | BASS_ENCODE_AUTOFREE, 0, 0);
#endif
	}
	if (!encoder) { // failed
		Error(@"Couldn't start encoding");
		BASS_ChannelStop(rchan);
		rchan = 0;
		return;
	}
	// setup cast
	server = [self.addressField.stringValue UTF8String];
	pass = [self.passwordField.stringValue UTF8String];
	name = [self.nameField.stringValue UTF8String];
	url = [self.urlField.stringValue UTF8String];
	genre = [self.genreField.stringValue UTF8String];
	desc = [self.descriptionField.stringValue UTF8String];
	flags = 0;
	if (self.sslSwitch.state) flags |= BASS_ENCODE_CAST_SSL;
	if (self.publicSwitch.state) flags |= BASS_ENCODE_CAST_PUBLIC;
	if (!BASS_Encode_CastInit(encoder, server, pass, content, name, url, genre, desc, NULL, bitrate, flags)) {
		Error(@"Couldn't setup connection with server");
		BASS_ChannelStop(rchan);
		rchan = 0;
		return;
	}
	BASS_ChannelPlay(rchan, FALSE); // resume recording
	self.startButton.title = @"Stop";
	[self.addressField setEnabled:false];
	[self.passwordField setEnabled:false];
	[self.sslSwitch setEnabled:false];
	[self.publicSwitch setEnabled:false];
	[self.nameField setEnabled:false];
	[self.urlField setEnabled:false];
	[self.genreField setEnabled:false];
	[self.descriptionField setEnabled:false];
	[self.codecSelector setEnabled:false];
	[self.bitrateSelector setEnabled:false];
	[self.titleField setEnabled:true];
	[self.sendButton setEnabled:true];

	BASS_Encode_SetNotify(encoder, EncoderNotify, (__bridge void*)self); // notify of dead encoder
}

- (void)Stop {
	// stop recording & encoding
	BASS_ChannelStop(rchan);
	rchan = 0;
	self.startButton.title = @"Start";
	[self.addressField setEnabled:true];
	[self.passwordField setEnabled:true];
	[self.sslSwitch setEnabled:true];
	[self.publicSwitch setEnabled:true];
	[self.nameField setEnabled:true];
	[self.urlField setEnabled:true];
	[self.genreField setEnabled:true];
	[self.descriptionField setEnabled:true];
	[self.codecSelector setEnabled:true];
	[self.bitrateSelector setEnabled:true];
	[self.titleField setEnabled:false];
	[self.sendButton setEnabled:false];
	self.view.window.title = @"Cast source example";
}

- (IBAction)startServer:(id)sender {
	if (!rchan)
		[self Start];
	else
		[self Stop];
}

- (IBAction)sendTitle:(id)sender {
	// send title update
	const char *title = [self.titleField.stringValue UTF8String];
	[sender setEnabled:false]; // temporarily disable button to prevent too frequent updates
#ifdef USE_ADDONS
	int codec = (int)self.codecSelector.indexOfSelectedItem;
	// Icecast requires Opus/FLAC title changes in new bitstreams
	if (codec == 2) { // Opus
		char title2[500];
		snprintf(title2, sizeof(title2) - 1, "--title \"%s\"", title);
		BASS_Encode_OPUS_NewStream(encoder, title2, 0);
	} else if (codec == 3) { // FLAC
		char title2[500];
		snprintf(title2, sizeof(title2) - 1, "-T \"title=%s\"", title);
		BASS_Encode_FLAC_NewStream(encoder, title2, 0);
	} else
#endif
		BASS_Encode_CastSetTitle(encoder, title, NULL);
	titlewait = 5;
}

- (void)TimerProc:(NSTimer*)timer {
	// update the level display
	float level = 0;
	if (rchan) {
		BASS_ChannelGetLevelEx(rchan, &level, 0.1, BASS_LEVEL_MONO); // get current level
		if (level > 0) {
			level = 1 + log10(level) * 20 / 30; // convert to dB (30dB range)
			if (level < 0) level = 0;
		}
	}
	self.levelIndicator.doubleValue = level;
	// handle auto-reconnect
	if (reconwait) {
		if (!--reconwait) {
			[self Stop];
			[self Start];
		}
	} else if (rchan) {
		// check number of listeners
		static int updatelisten = 0;
		if (!(++updatelisten % 50)) { // only checking once every 5 seconds
			const char *stats;
			int listeners = 0;
			if (stats = BASS_Encode_CastGetStats(encoder, BASS_ENCODE_STATS_ICE, NULL)) {
				char *t = strcasestr(stats, "<Listeners>"); // Icecast listener count
				if (t) listeners = atoi(t + 11);
			} else if (stats = BASS_Encode_CastGetStats(encoder, BASS_ENCODE_STATS_SHOUT, NULL)) {
				char *t = strstr(stats, "<CURRENTLISTENERS>"); // Shoutcast listener count
				if (t) listeners = atoi(t + 18);
			} else goto skiplisteners;
			self.view.window.title = [NSString stringWithFormat:@"Cast source example - %d listeners", listeners];
		}
skiplisteners:
		if (titlewait) {
			if (!--titlewait) [self.sendButton setEnabled:true];
		}
	}
}

- (void)viewDidLoad {
	[super viewDidLoad];

	// check the correct BASS was loaded
	if (HIWORD(BASS_GetVersion()) != BASSVERSION) {
		Error(@"An incorrect version of BASS was loaded");
		exit(0);
	}

	// initialize default recording device
	if (!BASS_RecordInit(-1)) {
		Error(@"Can't initialize device");
		exit(0);
	}

	{
		int c;
		for (c = 0; c < 14; c++)
			[self.bitrateSelector addItemWithTitle:[NSString stringWithFormat:@"%d",bitrates[c]]];
		[self.bitrateSelector selectItemAtIndex:8]; // default bitrate = 128kbps
	}

	[[NSRunLoop currentRunLoop] addTimer:[NSTimer timerWithTimeInterval:0.1 target:self selector:@selector(TimerProc:) userInfo:nil repeats:YES] forMode:NSRunLoopCommonModes];
}

@end
