/*
	BASSenc recording example
	Copyright (c) 2003-2020 Un4seen Developments Ltd.
*/

#import "ViewController.h"
#include "bass.h"
#include "bassenc.h"

#ifdef USE_ADDONS // use BASSenc add-ons instead of command-line encoders
// these headers should be placed alongside bassenc.h
#include "bassenc_mp3.h"
#include "bassenc_ogg.h"
#endif

@implementation ViewController

int input;		// current input source
int freq, chans;	// sample format
int codec;		// current encoder

HRECORD rchan;		// recording/encoding channel
HSTREAM pchan;		// playback channel
HENCODE encoder;	// encoder handle

// encoder output files
const char *files[] = { "bass.wav","bass.mp3","bass.ogg","bass.mp4","bass.m4a" };

// display error messages
void Error(NSString *es)
{
	es = [es stringByAppendingFormat:@"\n(error code: %d)",BASS_ErrorGetCode()];
	if (![NSThread isMainThread]) {
		dispatch_async(dispatch_get_main_queue(), ^{
			NSAlert *alert=[[NSAlert alloc] init];
			[alert setMessageText:es];
			[alert runModal];
		});
	} else {
		NSAlert *alert=[[NSAlert alloc] init];
		[alert setMessageText:es];
		[alert runModal];
	}
}

BOOL32 CALLBACK RecordingCallback(HRECORD handle, const void *buffer, DWORD length, void *user)
{
	return BASS_Encode_IsActive(handle); // continue recording if encoder is alive
}

- (void)Start {
	if (pchan) { // free old recording
		BASS_StreamFree(pchan);
		pchan = 0;
		[self.playButton setEnabled:false];
	}
	{ // get selected sample format
		int format = (int)self.formatSelector.indexOfSelectedItem;
		freq = format > 3 ? 22050 : format > 1 ? 44100 : 48000;
		chans = 1 + (format & 1);
	}
	// start recording (paused to add encoder first)
	rchan = BASS_RecordStart(freq, chans, BASS_RECORD_PAUSE, RecordingCallback, 0);
	if (!rchan) {
		Error(@"Couldn't start recording");
		return;
	}
	codec = (int)self.codecSelector.indexOfSelectedItem;
	// start the encoder
	if (!codec) { // PCM
		encoder = BASS_Encode_Start(rchan, files[codec], BASS_ENCODE_PCM, 0, 0);
	} else if (codec == 3) { // AAC
		encoder = BASS_Encode_StartCAFile(rchan, 'mp4f', 'aac ', 0, 0, files[codec]);
	} else if (codec == 4) { // ALAC
		encoder = BASS_Encode_StartCAFile(rchan, 'm4af', 'alac', 0, 0, files[codec]);
	} else {
#ifdef USE_ADDONS
		if (codec == 1) { // MP3
			encoder = BASS_Encode_MP3_StartFile(rchan, NULL, 0, files[codec]);
		} else { // Ogg Vorbis
			encoder = BASS_Encode_OGG_StartFile(rchan, NULL, 0, files[codec]);
		}
#else
		const char *com;
		if (codec == 1) { // MP3
			com = "lame - bass.mp3";
		} else { // Ogg Vorbis
			com = "oggenc -o bass.ogg -";
		}
		encoder = BASS_Encode_Start(rchan, com, 0, 0, 0);
#endif
	}
	if (!encoder) { // failed
		Error(@"Couldn't start encoding");
		BASS_ChannelStop(rchan);
		rchan = 0;
		return;
	}
	BASS_ChannelPlay(rchan, FALSE); // resume recoding
	self.recordButton.title = @"Stop";
	[self.formatSelector setEnabled:false];
	[self.codecSelector setEnabled:false];
}

- (void)Stop {
	// stop recording & encoding
	BASS_ChannelStop(rchan);
	rchan = 0;
	BASS_Encode_Stop(encoder);
	// create a stream from the recording
	if (pchan = BASS_StreamCreateFile(FALSE, files[codec], 0, 0, 0))
		[self.playButton setEnabled:true];
	self.recordButton.title = @"Record";
	[self.formatSelector setEnabled:true];
	[self.codecSelector setEnabled:true];
}

- (IBAction)startRecording:(id)sender {
	if (!rchan)
		[self Start];
	else
		[self Stop];
}

- (IBAction)startPlaying:(id)sender {
	BASS_ChannelPlay(pchan, TRUE); // play the recorded data
}

- (IBAction)changeDevice:(id)sender {
	// device selection changed
	int i = (int)[sender indexOfSelectedItem]; // get the selection
	if (rchan) rchan = 1; // special handle (real handles always have highest bit set) to prevent timer ending the recording
	// initialize the selected device
	if ([self InitDevice:i]) {
		if (rchan) { // continue recording on the new device...
			HRECORD newrchan = BASS_RecordStart(freq, chans, 0, RecordingCallback, 0);
			if (!newrchan)
				Error(@"Couldn't start recording");
			else {
				rchan = newrchan;
				BASS_Encode_SetChannel(encoder, rchan); // move the encoder to the new recording
			}
		}
	}
}

- (IBAction)changeInput:(id)sender {
	int i;
	input = (int)[sender indexOfSelectedItem]; // get the selection
	// enable the selected input
	for (i = 0; BASS_RecordSetInput(i, BASS_INPUT_OFF, -1); i++); // 1st disable all inputs, then...
	BASS_RecordSetInput(input, BASS_INPUT_ON, -1); // enable the selected
	[self UpdateInputInfo]; // update info
}

- (IBAction)changeVolume:(id)sender {
	double level = [sender doubleValue] / 100;
	if (!BASS_RecordSetInput(input, 0, level)) // failed to set input level
		BASS_RecordSetInput(-1, 0, level); // try master level instead
}

- (void)TimerProc:(NSTimer*)timer {
	// update the level display
	float level = 0;
	if (rchan || pchan) {
		BASS_ChannelGetLevelEx(rchan ? rchan : pchan, &level, 0.1, BASS_LEVEL_MONO); // get current level
		if (level > 0) {
			level = 1 + log10(level) * 20 / 30; // convert to dB (30dB range)
			if (level < 0) level = 0;
		}
	}
	self.levelIndicator.doubleValue = level;
	// update the recording/playback counter
	char text[30] = "";
	if (rchan) { // recording/encoding
		if (rchan != 1 && !BASS_ChannelIsActive(rchan)) { // the recording has stopped, eg. unplugged device
			[self Stop];
			Error(@"The recording stopped");
		} else
			sprintf(text, "%lld", BASS_Encode_GetCount(encoder, BASS_ENCODE_COUNT_IN));
	} else if (pchan) {
		if (BASS_ChannelIsActive(pchan)) // playing
			sprintf(text, "%lld / %lld", BASS_ChannelGetPosition(pchan, BASS_POS_BYTE), BASS_ChannelGetLength(pchan, BASS_POS_BYTE));
		else
			sprintf(text, "%lld", BASS_ChannelGetLength(pchan, BASS_POS_BYTE));
	}
	self.positionText.stringValue=[NSString stringWithUTF8String:text];
}

- (void)UpdateInputInfo {
	float level;
	int it = BASS_RecordGetInput(input, &level); // get info on the input
	if (it == -1 || level < 0) { // failed to get level
		it = BASS_RecordGetInput(-1, &level); // try master input instead
		if (it == -1 || level < 0) { // that failed too
			level = 1; // just display 100%
			[self.volumeSlider setEnabled:false];
		} else
			[self.volumeSlider setEnabled:true];
	} else
		[self.volumeSlider setEnabled:true];
	self.volumeSlider.doubleValue = level * 100; // set the level slider
}

- (BOOL)InitDevice:(int)device {
	BASS_RecordFree(); // free current device (and recording channel) if there is one
	// initalize new device
	if (!BASS_RecordInit(device)) {
		Error(@"Can't initialize recording device");
		return FALSE;
	}
	{ // get list of inputs
		int c;
		const char *name;
		[self.inputSelector removeAllItems];
		for (c = 0; name = BASS_RecordGetInputName(c); c++) {
			[self.inputSelector addItemWithTitle:[NSString stringWithUTF8String:name]];
			if (!(BASS_RecordGetInput(c, NULL) & BASS_INPUT_OFF)) { // this one is currently "on"
				input = c;
				[self.inputSelector selectItemAtIndex:c];
			}
		}
		if (!c && (name = BASS_RecordGetInputName(-1))) {
			[self.inputSelector addItemWithTitle:[NSString stringWithUTF8String:name]];
			[self.inputSelector selectItemAtIndex:0];
		}
		[self UpdateInputInfo]; // display info
	}
	return TRUE;
}


- (void)viewDidLoad {
	[super viewDidLoad];

	// check the correct BASS was loaded
	if (HIWORD(BASS_GetVersion()) != BASSVERSION) {
		Error(@"An incorrect version of BASS was loaded");
		exit(0);
	}

	// initialize default output device
	if (!BASS_Init(-1, 44100, 0, NULL, NULL))
		Error(@"Can't initialize output device");

	{ // get list of recording devices
		int c, def = 0;
		BASS_DEVICEINFO di;
		for (c = 0; BASS_RecordGetDeviceInfo(c, &di); c++) {
			[self.deviceSelector addItemWithTitle:[NSString stringWithUTF8String:di.name]];
			if (di.flags & BASS_DEVICE_DEFAULT) { // got the default device
				[self.deviceSelector selectItemAtIndex:c];
				def = c;
			}
		}
		[self InitDevice:def]; // initialize default recording device
	}

	[[NSRunLoop currentRunLoop] addTimer:[NSTimer timerWithTimeInterval:0.1 target:self selector:@selector(TimerProc:) userInfo:nil repeats:YES] forMode:NSRunLoopCommonModes];
}

@end
