/*
	BASSenc server example
	Copyright (c) 2011-2022 Un4seen Developments Ltd.
*/

#import "ViewController.h"
#include "bass.h"
#include "bassenc.h"

#ifdef USE_ADDONS // use BASSenc add-ons instead of command-line encoders
// these headers should be placed alongside bassenc.h
#include "bassenc_mp3.h"
#include "bassenc_ogg.h"
#include "bassenc_opus.h"
#include "bassenc_flac.h"
#endif

@implementation ViewController

HRECORD rchan;		// recording/encoding channel
HENCODE encoder;	// encoder handle
char icyheads[1000];
int titlewait;

const DWORD bitrates[14] = { 32,40,48,56,64,80,96,112,128,160,192,224,256,320 }; // available bitrates

// display error messages
void Error(NSString *es)
{
	es = [es stringByAppendingFormat:@"\n(error code: %d)",BASS_ErrorGetCode()];
	if (![NSThread isMainThread]) {
		dispatch_async(dispatch_get_main_queue(), ^{
			NSAlert *alert=[[NSAlert alloc] init];
			[alert setMessageText:es];
			[alert runModal];
		});
	} else {
		NSAlert *alert=[[NSAlert alloc] init];
		[alert setMessageText:es];
		[alert runModal];
	}
}

BOOL32 CALLBACK RecordingCallback(HRECORD handle, const void *buffer, DWORD length, void *user)
{
	return BASS_Encode_IsActive(handle); // continue recording if encoder is alive
}

// encoder death notification
void CALLBACK EncoderNotify(HENCODE handle, DWORD status, void *user)
{
	if (status < 0x10000) { // encoder died
		dispatch_async(dispatch_get_main_queue(), ^{
			ViewController *view=(__bridge ViewController*)user;
			[view Stop]; // free the recording and encoder
			NSAlert *alert=[[NSAlert alloc] init];
			[alert setMessageText:@"The encoder died!"];
			[alert runModal];
		});
	}
}

// client connection/disconnection notification
BOOL32 CALLBACK EncodeClientProc(HENCODE handle, BOOL32 connect, const char *client, char *headers, void *user)
{
	if (connect) strcpy(headers, icyheads); // set response headers
	dispatch_async(dispatch_get_main_queue(), ^{
		ViewController *view=(__bridge ViewController*)user;
		if (connect) { // add to client list
			[view.clientListController addObject:[NSString stringWithUTF8String:client]];
		} else { // disconnecting, remove from client list
			[view.clientListController removeObject:[NSString stringWithUTF8String:client]];
		}
	});
	return TRUE; // allow connectiom
}

- (void)Start {
	char buf[200];
	const char *server;
	DWORD codec, bitrate, port, flags;
	// start recording @ 44100hz 16-bit stereo (paused to setup encoder first)
	if (!(rchan = BASS_RecordStart(44100, 2, BASS_RECORD_PAUSE, RecordingCallback, 0))) {
		Error(@"Couldn't start recording");
		return;
	}
	flags = (self.httpsSwitch.state ? BASS_ENCODE_SERVER_SSL : 0);
	codec = (int)self.codecSelector.indexOfSelectedItem;
	bitrate = bitrates[(int)self.bitrateSelector.indexOfSelectedItem]; // get bitrate
	// start the encoder
	if (codec == 4) { // AAC
		encoder = BASS_Encode_StartCA(rchan, 'adts', 'aac ', BASS_ENCODE_AUTOFREE, bitrate * 1000, 0, 0);
		flags |= BASS_ENCODE_SERVER_META;
	} else if (codec == 5) { // PCM
		encoder = BASS_Encode_Start(rchan, NULL, BASS_ENCODE_PCM | BASS_ENCODE_AUTOFREE, 0, 0);
		bitrate = (DWORD)BASS_ChannelSeconds2Bytes(rchan, 1) / 125;
		flags |= BASS_ENCODE_SERVER_META;
	} else {
#ifdef USE_ADDONS
		if (!codec) { // MP3
			sprintf(buf, "-b %d", bitrate);
			encoder = BASS_Encode_MP3_Start(rchan, buf, BASS_ENCODE_AUTOFREE, 0, 0);
			flags |= BASS_ENCODE_SERVER_META;
		} else if (codec == 1) { // Ogg Vorbis
			sprintf(buf, "-M %d -m %d", bitrate, bitrate);
			encoder = BASS_Encode_OGG_Start(rchan, buf, BASS_ENCODE_AUTOFREE, 0, 0);
		} else if (codec == 2) { // Opus
			sprintf(buf, "--bitrate %d --cvbr --max-delay 100 --padding 0", bitrate);
			encoder = BASS_Encode_OPUS_Start(rchan, buf, BASS_ENCODE_AUTOFREE, 0, 0);
		} else { // FLAC
			sprintf(buf, "--ogg --no-padding --no-seektable --limit-min-bitrate");
			encoder = BASS_Encode_FLAC_Start(rchan, buf, BASS_ENCODE_AUTOFREE, 0, 0);
			bitrate = (DWORD)BASS_ChannelSeconds2Bytes(rchan, 0.5) / 125; // assuming 50% reduction
		}
#else
		// setup encoder command-line (raw PCM data to avoid length limit)
		if (!codec) { // MP3
			sprintf(buf, "lame -b %d --flush -r -s 44100 -", bitrate); // add "-x" for LAME versions pre-3.98
		} else if (codec == 1) { // Ogg Vorbis
			sprintf(buf, "oggenc -M %d -m %d -r -R 44100 -", bitrate, bitrate);
		} else if (codec == 2) { // Opus
			sprintf(buf, "opusenc --bitrate %d --cvbr --max-delay 100 --padding 0 --raw --raw-rate 44100 - -", bitrate);
		} else { // FLAC
			sprintf(buf, "flac --ogg --no-padding --no-seektable --limit-min-bitrate --sample-rate=44100 --channels=2 --bps=16 --endian=little --sign=signed -");
			bitrate = (DWORD)BASS_ChannelSeconds2Bytes(rchan, 0.5) / 125; // assuming 50% reduction
		}
		encoder = BASS_Encode_Start(rchan, buf, BASS_ENCODE_NOHEAD | BASS_ENCODE_AUTOFREE, 0, 0);
		flags |= BASS_ENCODE_SERVER_META;
#endif
	}
	if (!encoder) { // failed
		Error(@"Couldn't start encoding");
		BASS_ChannelStop(rchan);
		rchan = 0;
		return;
	}
	server = [self.portField.stringValue UTF8String];
	// start the server with 5 second buffer (burst the full amount)
	port = BASS_Encode_ServerInit(encoder, server, bitrate * 125 * 5, -1, flags, EncodeClientProc, (__bridge void*)self);
	if (!port) { // failed
		if (BASS_ErrorGetCode() == BASS_ERROR_SERVER_CERT)
			Error(@"Couldn't start server due to a certificate issue");
		else
			Error(@"Couldn't start server");
		BASS_ChannelStop(rchan);
		rchan = 0;
		return;
	}
	{ // build stream info headers for clients
		const char *name, *url, *genre;
		char *p = icyheads;
		*p = 0;
		name = [self.nameField.stringValue UTF8String];
		if (strlen(name)) p += sprintf(p, "icy-name:%s\r\n", name);
		url = [self.urlField.stringValue UTF8String];
		if (strlen(url)) p += sprintf(p, "icy-url:%s\r\n", url);
		genre = [self.genreField.stringValue UTF8String];
		if (strlen(genre)) p += sprintf(p, "icy-genre:%s\r\n", genre);
	}
	BASS_ChannelPlay(rchan, FALSE); // resume recording
	{ // update port number in display
		char newbuf[210], *path, *p;
		strncpy(buf, server, sizeof(buf) - 1);
		path = strchr(buf, '/');
		if (path) *path++ = 0;
		if ((p = strpbrk(buf, ".]"))) { // got an IP address
			if ((p = strchr(p, ':'))) *p = 0;
			sprintf(newbuf, path ? "%s:%d/%s" : "%s:%d", buf, port, path);
		} else
			sprintf(newbuf, path ? "%d/%s" : "%d", port, path);
		self.portField.stringValue = [NSString stringWithUTF8String:newbuf];
	}
	self.startButton.title = @"Stop";
	[self.portField setEnabled:false];
	[self.httpsSwitch setEnabled:false];
	[self.nameField setEnabled:false];
	[self.urlField setEnabled:false];
	[self.genreField setEnabled:false];
	[self.codecSelector setEnabled:false];
	[self.bitrateSelector setEnabled:false];
	[self.titleField setEnabled:true];
	[self.sendButton setEnabled:true];
	[self.kickButton setEnabled:true];

	BASS_Encode_SetNotify(encoder, EncoderNotify, (__bridge void*)self); // notify of dead encoder
}

- (void)Stop {
	// stop recording & encoding
	BASS_ChannelStop(rchan);
	rchan = 0;
	self.startButton.title = @"Start";
	[self.portField setEnabled:true];
	[self.httpsSwitch setEnabled:true];
	[self.nameField setEnabled:true];
	[self.urlField setEnabled:true];
	[self.genreField setEnabled:true];
	[self.codecSelector setEnabled:true];
	[self.bitrateSelector setEnabled:true];
	[self.titleField setEnabled:false];
	[self.sendButton setEnabled:false];
	[self.kickButton setEnabled:false];
	[self.clientListController setContent:nil];
}

- (IBAction)startServer:(id)sender {
	if (!rchan)
		[self Start];
	else
		[self Stop];
}

- (IBAction)sendTitle:(id)sender {
	// send title update
	const char *title = [self.titleField.stringValue UTF8String];
	[sender setEnabled:false]; // temporarily disable button to prevent too frequent updates
#ifdef USE_ADDONS
	int codec = (int)self.codecSelector.indexOfSelectedItem;
	// use a new bitstream for title changes with Ogg-based formats
	if (codec == 1) { // Ogg Vorbis
		char title2[500];
		sprintf(title2, "--title \"%s\"", title);
		BASS_Encode_OGG_NewStream(encoder, title2, 0);
	} else if (codec == 2) { // Opus
		char title2[500];
		sprintf(title2, "--title \"%s\"", title);
		BASS_Encode_OPUS_NewStream(encoder, title2, 0);
	} else if (codec == 3) { // FLAC
		char title2[500];
		sprintf(title2, "-T \"title=%s\"", title);
		BASS_Encode_FLAC_NewStream(encoder, title2, 0);
	} else
#endif
		BASS_Encode_CastSetTitle(encoder, title, NULL);
	titlewait = 5;
}

- (IBAction)kickClient:(id)sender {
	for (id ip in self.clientListController.selectedObjects)
		BASS_Encode_ServerKick(encoder, [ip UTF8String]);
}

- (void)TimerProc:(NSTimer*)timer {
	// update the level display
	float level = 0;
	if (rchan) {
		BASS_ChannelGetLevelEx(rchan, &level, 0.1, BASS_LEVEL_MONO); // get current level
		if (level > 0) {
			level = 1 + log10(level) * 20 / 30; // convert to dB (30dB range)
			if (level < 0) level = 0;
		}
	}
	self.levelIndicator.doubleValue = level;
	if (rchan && titlewait) {
		if (!--titlewait) [self.sendButton setEnabled:true];
	}
}

- (void)viewDidLoad {
	[super viewDidLoad];

	// example SSL certificate for HTTPS
	BASS_SetConfigPtr(BASS_CONFIG_ENCODE_SERVER_CERT, [[[NSBundle mainBundle] pathForResource:@"testcert" ofType:@"crt"] UTF8String]);
	BASS_SetConfigPtr(BASS_CONFIG_ENCODE_SERVER_KEY, [[[NSBundle mainBundle] pathForResource:@"testcert" ofType:@"key"] UTF8String]);

	// check the correct BASS was loaded
	if (HIWORD(BASS_GetVersion()) != BASSVERSION) {
		Error(@"An incorrect version of BASS was loaded");
		exit(0);
	}

	// initialize default recording device
	if (!BASS_RecordInit(-1)) {
		Error(@"Can't initialize device");
		exit(0);
	}

	{
		int c;
		for (c = 0; c < 14; c++)
			[self.bitrateSelector addItemWithTitle:[NSString stringWithFormat:@"%d",bitrates[c]]];
		[self.bitrateSelector selectItemAtIndex:8]; // default bitrate = 128kbps
	}

	[[NSRunLoop currentRunLoop] addTimer:[NSTimer timerWithTimeInterval:0.1 target:self selector:@selector(TimerProc:) userInfo:nil repeats:YES] forMode:NSRunLoopCommonModes];
}

@end
